/* 
	In-place Transform

	compile with:

	nvcc -o cuda.o -shared cudatest.cu -I/usr/local/cuda/NVIDIA_CUDA_SDK/common/inc -I/usr/local/cuda -Xcompiler -fPIC
	gcc -o cuda.so -shared /home/ancon1sm/voltools/gpgpu/cuda.o -fPIC -L/usr/local/cuda/NVIDIA_CUDA_SDK/lib -L/usr/local/cuda/lib -lcudart -lcufft
 
*/

#include <python2.4/Python.h>
#include <python2.4/numpy/arrayobject.h>
#include <cutil.h>
#include <cufft.h>

typedef float2 Complex;

float ***PyArray_to_FCArray(PyArrayObject *arrayin);
Complex ***PyArray_to_CCArray(PyArrayObject *arrayin);
cufftComplex *PyArray_to_CuArray(PyArrayObject *arrayin);

float ***
PyArray_to_FCArray(PyArrayObject *arrayin)
{
	float ***arrayout, *a;
	int i, j, m, n, o;

	m = arrayin->dimensions[0];
	n = arrayin->dimensions[1];
	o = arrayin->dimensions[2];
	a = (float *) arrayin->data;

	arrayout = (float ***)malloc(sizeof(float **) *m);
	for (i = 0; i < m; i++)
	{
		arrayout[i] = (float **)malloc(sizeof(float *) *n);
		for (j = 0; j < n; j++)
			arrayout[i][j] = a + i*m*o + j*o;
	}
	return arrayout;
}

Complex ***
PyArray_to_CCArray(PyArrayObject *arrayin)
{
	Complex ***arrayout, *a;
	int i, j, m, n, o;

	m = arrayin->dimensions[0];
	n = arrayin->dimensions[1];
	o = arrayin->dimensions[2];
	a = (Complex *) arrayin->data;

	arrayout = (Complex ***)malloc(sizeof(Complex **) *m);
	for (i = 0; i < m; i++)
	{
		arrayout[i] = (Complex **)malloc(sizeof(Complex *) *n);
		for (j = 0; j < n; j++)
			arrayout[i][j] = a + i*m*o + j*o;
	}
	return arrayout;
}

cufftComplex *
FPyArray_to_CuArray(PyArrayObject *arrayin)
{
	cufftComplex *arrayout, *cudaarray;
	float ***arraytemp;
	int i, j, k, m, n, o;

	m = arrayin->dimensions[0];
	n = arrayin->dimensions[1];
	o = arrayin->dimensions[2];

	arraytemp = PyArray_to_FCArray(arrayin);
	arrayout = (cufftComplex *)malloc(sizeof(cufftComplex)*m*n*o);

		for (i = 0; i < m; i++)
		for (j = 0; j < n; j++)
			for (k = 0; k < o; k++)
			{
				arrayout[i*m + j*n + k].x = arraytemp[i][j][k];
				arrayout[i*m + j*n + k].y = 0.0;
			}

	cudaMalloc((void**)&cudaarray, sizeof(cufftComplex)*m*n*o);
	cudaMemcpy(cudaarray, arrayout, sizeof(cufftComplex)*m*n*o, cudaMemcpyHostToDevice);
	return cudaarray;
}

cufftComplex *
CPyArray_to_CuArray(PyArrayObject *arrayin)
{
	cufftComplex *arrayout, *cudaarray;
	Complex ***arraytemp;
	int i, j, k, m, n, o;

	m = arrayin->dimensions[0];
	n = arrayin->dimensions[1];
	o = arrayin->dimensions[2];

	arraytemp = PyArray_to_CCArray(arrayin);
	arrayout = (cufftComplex *)malloc(sizeof(cufftComplex)*m*n*o);

	for (i = 0; i < m; i++)
		for (j = 0; j < n; j++)
			for (k = 0; k < o; k++)
			{
				arrayout[i*m + j*n + k].x = arraytemp[i][j][k].x;
				arrayout[i*m + j*n + k].y = arraytemp[i][j][k].y;
			}
			
	return cudaarray;
}

cufftComplex ***
CPyArray_to_CArray(PyArrayObject *arrayin)
{
	cufftComplex ***arrayout, *a;
	int i, j, m, n, o;

	m = arrayin->dimensions[0];
	n = arrayin->dimensions[1];
	o = arrayin->dimensions[2];
	a = (cufftComplex *) arrayin->data;
	
	arrayout = (cufftComplex ***)malloc(sizeof(cufftComplex **) * m);
	for (i = 0; i < m; i++)
	{
		arrayout[i] = (cufftComplex **)malloc(sizeof(cufftComplex *) * n);
		for (j = 0; j < n; j++)
			arrayout[i][j] = a + i*m*o + j*o;
	}
	return arrayout;
}

static PyObject*
test(PyObject* self, PyObject* args)
{
	if (!PyArg_ParseTuple(args, ""))
		return 0;
	return Py_BuildValue("s", "Cuda works!");
}

extern "C"
static PyObject*
fft3(PyObject* self, PyObject* args)
{
	PyObject *input;
	PyArrayObject *a, *b;
	cufftComplex *cin, *cout, *cudaarray;
	cufftHandle plan;
	int m, n, o, dim[3];

	if (!PyArg_ParseTuple(args, "O", &input))
		return NULL;

	a = (PyArrayObject *)PyArray_ContiguousFromObject(input, NPY_CFLOATLTR, 3, 3);
	
	m = dim[0] = a->dimensions[0];
	n = dim[1] = a->dimensions[1];
	o = dim[2] = a->dimensions[2];
	b = (PyArrayObject* ) PyArray_FromDims(3, dim, NPY_CFLOATLTR);
	
	// Convert PyArrayObjects to C pointers
	cin = CPyArray_to_CuArray(a);
	cout = (cufftComplex *)malloc(sizeof(cufftComplex)*m*n*o);
	cout = (cufftComplex *) b->data;
	//cout = CPyArray_to_CArray(b);

	//Add cuda stuff here
	cufftPlan3d(&plan, m, n, o, CUFFT_C2C);
	cudaMalloc((void**)&cudaarray, sizeof(cufftComplex)*m*n*o);
	cudaMemcpy(cudaarray, cin, sizeof(cufftComplex)*m*n*o, cudaMemcpyHostToDevice);
	cufftExecC2C(plan, cin, cin, CUFFT_FORWARD);
	cudaMemcpy(cout, cin, sizeof(cufftComplex)*m*n*o, cudaMemcpyDeviceToHost);
	cufftDestroy(plan);
	cudaFree(cin);
	cudaFree(cout);

	//fprintf(stderr, "Testing output.\n");
	//int i, j, k;
	//for (i = 0; i < m; i++)
	//	for (j = 0; j < n; j++)
	//		for (k = 0; k < o; k++)
	//			fprintf(stderr, "%f + %fj\n", cout[i*m + j*n +k].x, cout[i*m + j*n +k].y);

	return PyArray_Return(b);
}

static PyMethodDef cudaMethods[] = {
	{"test", test, METH_VARARGS, "Cuda test."},
	{"fft3", fft3, METH_VARARGS, "Cuda 3D FFT test."},
	{NULL, NULL, 0, NULL}
};

extern "C"
void initcuda()
{
	Py_InitModule("cuda", cudaMethods);
	import_array(); //Needed for Numpy
}
